@@ -20,6 +20,7 @@ from PyQt4 import QtCore
 from PyKDE4 import soprano
 import sys
 import codecs
+import glob
 
 
 # our tree, maps URIs to EntityContainer objects
@@ -52,6 +53,15 @@ def elidedText(s, maxLen):
         return s.left(maxLen) + "..."
 
 
+def resolveAbbreviatedUri(abbrevUri):
+  abbr = abbrevUri.left(abbrevUri.indexOf(':'))
+  name = abbrevUri.mid(abbr.length()+1)
+  for ns in allNamespaces.keys():
+    if abbr == allNamespaces[ns]['abbr']:
+      return ns + name
+  return ''
+
+
 def buildDocBookEntity(entityUri):
   if entityUri in entityTree:
     return entityTree[entityUri].createDocBookLink()
@@ -69,6 +79,13 @@ def buildDocBookEntityList(entityUris, directRelations = ()):
   return s.join(', ')
 
 
+def buildXRefList(linkends):
+  s = QtCore.QStringList()
+  for link in linkends:
+    s.append('<xref linkend="%s" />' % link)
+  return s.join(', ')
+
+
 class EntityContainer(object):
   """
   A container for a class or a property with members
@@ -131,7 +148,11 @@ class EntityContainer(object):
 
   def toDocBook(self):
     entityId = "%s:%s" % (allNamespaces[self.namespace()]["abbr"], self.name())
-    s = '<section id="%s"><title>%s</title><informaltable>' % (entityId, entityId)
+    s = '<section id="%s"><title>' % entityId
+    s += entityId
+    if self.deprecated:
+      s += ' <emphasis>(deprecated)</emphasis>'
+    s += '</title><informaltable>'
     s += '<tgroup cols="2"><tbody>'
     s += '<row><entry align="right">Label</entry><entry>%s</entry></row>' % self.label
     s += '<row><entry align="right">Description</entry><entry><![CDATA[%s]]></entry></row>' % self.comment()
@@ -157,6 +178,8 @@ class EntityContainer(object):
       s += '<row><entry align="right">In range of</entry><entry>%s</entry></row>' % buildDocBookEntityList(self.inRangeOf)
       if len(self.instances) > 0:
         s += '<row><entry align="right">Instances</entry><entry>%s</entry></row>' % buildDocBookEntityList(self.instances)
+    if len(self.backlinks) > 0:
+      s += '<row><entry align="right">Mentioned in</entry><entry>%s</entry></row>' % buildXRefList(self.backlinks)
     s += '</tbody></tgroup></informaltable></section>'
     return s
 
@@ -181,6 +204,8 @@ class EntityContainer(object):
     self.maxCardinality = -1
     self.m_comment = QtCore.QString()
     self.inverseProperty = QtCore.QString();
+    self.backlinks = set()
+    self.deprecated = False
 
   def __cmp__(self, other):
     if self.name() < other.name(): return -1
@@ -220,6 +245,13 @@ def buildEntityTree(ontologyFiles):
     if uri in entityTree:
       entityTree[uri].m_comment = it.current().object().toString()
 
+  # mark deprecated classes and properties
+  it = graph.listStatements(soprano.Soprano.Node(), soprano.Soprano.Node(soprano.Soprano.Vocabulary.NAO.deprecated()), soprano.Soprano.Node())
+  while(it.next()):
+    uri = it.current().subject().uri().toString()
+    if uri in entityTree:
+      entityTree[uri].deprecated = it.current().object().literal().toBool()
+
   # add hierarchy
   subRelations = {}
   superRelations = {}
@@ -311,6 +343,39 @@ def buildEntityTree(ontologyFiles):
       allNamespaces.setdefault(ns, {})["desc"] = it.current().object().literal().toString()
 
 
+def extractEntityBacklinks(docbookFolder):
+  "Extracts links from documentation to classes and properties. The backlinks are stored in the entityTree as string ids to be used in xref docbook links."
+  # Parse each docbook file and find xref links which relate to a class or a property
+  # for each xref find the containing section and store its id in the entity's backlinks list
+  for filename in glob.glob(docbookFolder + '/*-main.docbook'):
+    #
+    # We need a well-formed docbook document for the stream reader.
+    # Thus we fake one by putting the file contents into an article
+    #
+    file = QtCore.QFile(filename)
+    file.open(QtCore.QIODevice.ReadOnly)
+    content = file.readAll()
+    content.prepend('<article>')
+    content.append('</article>')
+
+    xml = QtCore.QXmlStreamReader(content)
+    sectionStack = []
+    while not xml.atEnd():
+      tokenType = xml.readNext()
+      if tokenType == QtCore.QXmlStreamReader.StartElement:
+        if xml.name() == 'section':
+          if xml.attributes().hasAttribute('id'):
+            sectionStack.append(xml.attributes().value('id').toString())
+        elif xml.name() == 'xref':
+          linkend = xml.attributes().value('linkend').toString()
+          resolvedUri = resolveAbbreviatedUri(linkend)
+          if resolvedUri in entityTree:
+            entityTree[resolvedUri].backlinks.add(sectionStack[-1])
+      elif tokenType == QtCore.QXmlStreamReader.EndElement:
+        if xml.name() == 'section':
+          sectionStack.pop()
+
+
 def printEntityTree():
   for e in entityTree.values():
     print "%s (%s) (%s)" % (e.label, e.uri, e.isProperty)
@@ -352,14 +417,20 @@ def writeDocBookAppendix(ns):
   filename = '%s-reference.docbook' % name
   f = codecs.open(filename, 'w', 'utf-8')
   f.write('<section xmlns="http://docbook.org/ns/docbook"><title>%s Vocabulary Summary</title>' % name.toUpper())
-  f.write('<section><title>Description of Classes</title>')
-  for a in getSortedClasses(ns):
-    f.write(a.toDocBook())
-  f.write('</section>')
-  f.write('<section><title>Description of Properties</title>')
-  for a in getSortedProperties(ns):
-    f.write(a.toDocBook())
-  f.write('</section>')
+
+  sortedClasses = getSortedClasses(ns)
+  if len(sortedClasses) > 0:
+      f.write('<section><title>Description of Classes</title>')
+      for a in sortedClasses:
+          f.write(a.toDocBook())
+      f.write('</section>')
+
+  sortedProperties = getSortedProperties(ns)
+  if len(sortedProperties) > 0:
+      f.write('<section><title>Description of Properties</title>')
+      for a in sortedProperties:
+          f.write(a.toDocBook())
+      f.write('</section>')
   f.write('</section>')
 
 
@@ -386,20 +457,26 @@ def writeOntologyOverviewFile(ns):
     name = allNamespaces[ns]["abbr"]
     filename = '%s-overview.docbook' % name
     f = open(filename, 'w')
-    f.write('<section xmlns="http://docbook.org/ns/docbook">')
-    f.write('<title>Classes Overview</title>')
-    f.write('<informaltable><tgroup cols="1"><tbody>')
-    for a in getSortedClasses(ns):
-        f.write('<row><entry>%s</entry></row>' % a.toShortDocBook())
-    f.write('</tbody></tgroup></informaltable>')
-    f.write('</section>')
-    f.write('<section xmlns="http://docbook.org/ns/docbook">')
-    f.write('<title>Properties Overview</title>')
-    f.write('<informaltable><tgroup cols="1"><tbody>')
-    for a in getSortedProperties(ns):
-        f.write('<row><entry>%s</entry></row>' % a.toShortDocBook())
-    f.write('</tbody></tgroup></informaltable>')
-    f.write('</section>')
+
+    sortedClasses = getSortedClasses(ns)
+    if len(sortedClasses) > 0:
+        f.write('<section xmlns="http://docbook.org/ns/docbook">')
+        f.write('<title>Classes Overview</title>')
+        f.write('<informaltable><tgroup cols="1"><tbody>')
+        for a in sortedClasses:
+            f.write('<row><entry>%s</entry></row>' % a.toShortDocBook())
+        f.write('</tbody></tgroup></informaltable>')
+        f.write('</section>')
+
+    sortedProperties = getSortedProperties(ns)
+    if len(sortedProperties) > 0:
+        f.write('<section xmlns="http://docbook.org/ns/docbook">')
+        f.write('<title>Properties Overview</title>')
+        f.write('<informaltable><tgroup cols="1"><tbody>')
+        for a in getSortedProperties(ns):
+            f.write('<row><entry>%s</entry></row>' % a.toShortDocBook())
+        f.write('</tbody></tgroup></informaltable>')
+        f.write('</section>')
 
 def writeOntologyToc():
     "Writes a TOC file which lists links to all ontologies"
@@ -415,7 +492,7 @@ def writeOntologyToc():
 def writeClassIndex():
     "Writes one docbook file containing an index of all classes"
     f = open('classindex.docbook', 'w')
-    f.write('<section xmlns="http://docbook.org/ns/docbook">')
+    f.write('<chapter id="classindex" xmlns="http://docbook.org/ns/docbook">')
     f.write('<title>All Classes</title>')
     currentChar = ''
     for a in getSortedClasses():
@@ -429,14 +506,14 @@ def writeClassIndex():
             f.write('<informaltable><tgroup cols="1"><tbody>')
         f.write('<row><entry>%s</entry></row>' % a.toShortDocBook())
     f.write('</tbody></tgroup></informaltable></section>')
-    f.write('</section>')
+    f.write('</chapter>')
     f.close()
 
 
 def writePropertyIndex():
     "Writes one docbook file containing an index of all properties"
     f = open('propertyindex.docbook', 'w')
-    f.write('<section xmlns="http://docbook.org/ns/docbook">')
+    f.write('<chapter id="propertyindex" xmlns="http://docbook.org/ns/docbook">')
     f.write('<title>All Properties</title>')
     currentChar = ''
     for a in getSortedProperties():
@@ -450,12 +527,13 @@ def writePropertyIndex():
             f.write('<informaltable><tgroup cols="1"><tbody>')
         f.write('<row><entry>%s</entry></row>' % a.toShortDocBook())
     f.write('</tbody></tgroup></informaltable></section>')
-    f.write('</section>')
+    f.write('</chapter>')
     f.close()
 
 
 def main():
-  buildEntityTree(sys.argv[1:])
+  buildEntityTree(sys.argv[2:])
+  extractEntityBacklinks(sys.argv[1])
   # write an index of all classes
   writeClassIndex()
   writePropertyIndex()
